import configparser
import json
import multiprocessing.pool
import os
import random
import time
import logging
from datetime import datetime, timedelta
from functools import partial
from pathlib import Path
from typing import Callable, Any, List, NamedTuple, Union, Mapping

import boxsdk.exception
import dateutil
from boxsdk import Client, OAuth2
from boxsdk.exception import BoxAPIException
from boxsdk.object.event import Event
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from dateutil.parser import parse
from requests.exceptions import ConnectionError
from urllib3.exceptions import ProtocolError

from .cache_utils import (
    redis_key,
    redis_set,
    redis_get,
    r_c,
    local_or_box_file_m_time_key_func,
)
from .gui import notify_user_with_gui
from .log_utils import setup_logger
from .oauth_utils import setup_remote_oauth

setup_logger()

crate_logger = logging.getLogger(__name__)


class DownloadQueueItem(NamedTuple):
    item: Union[File, Folder]
    path: Union[Path, str]
    oauth: OAuth2
    event: Union[Event, Mapping, None]


class UploadQueueItemReal(NamedTuple):
    timestamp: float
    callable_up: Callable[..., Any]
    oauth: OAuth2
    explicit_file_path: Union[Path, str, None]


UploadQueueItem = Union[Callable[..., Any], UploadQueueItemReal]

uploads_given_up_on: List[Callable[..., Any]] = []


def upload_queue_processor(queue_item: UploadQueueItem):
    """
    Implements a simple re-try mechanism for pending uploads
    :return:
    """
    callable_up = queue_item
    # TODO: pass in the actual item being updated/uploaded,
    #  so we can do more intelligent retry mechanisms
    was_list = isinstance(callable_up, (list, tuple))
    last_modified_time = oauth = explicit_file_path = None
    if was_list:
        original_callable_up = callable_up
        last_modified_time, callable_up, oauth = original_callable_up[:3]
        if len(original_callable_up) > 3:
            explicit_file_path = original_callable_up[3]
        elif isinstance(callable_up, partial):
            explicit_file_path = callable_up.args[0]
    args = callable_up.args if isinstance(callable_up, partial) else None
    num_retries = 15
    perform_upload(
        args,
        callable_up,
        last_modified_time,
        num_retries,
        oauth,
        explicit_file_path,
        was_list,
        retry_limit=num_retries,
    )


def perform_upload(
    args,
    callable_up,
    last_modified_time,
    num_retries,
    oauth,
    explicit_file_path,
    was_list,
    retry_limit=15,
):
    for x in range(retry_limit):
        try:
            crate_logger.debug(
                f"Attempt upload{' ' + explicit_file_path if was_list else ''}"
            )
            ret_val = callable_up()
            crate_logger.info(
                f"Completed upload{' ' + explicit_file_path if was_list else ''}"
            )
            if was_list:
                path_name = explicit_file_path
                item = ret_val  # is the new/updated item
                if isinstance(item, File):
                    client = Client(oauth)
                    file_obj: File = client.file(file_id=item.object_id).get()
                    redis_set(r_c, file_obj, last_modified_time, box_dir_path=BOX_DIR)
                    r_c.set(
                        local_or_box_file_m_time_key_func(path_name, False),
                        datetime.fromtimestamp(Path(path_name).stat().st_mtime)
                        .astimezone(dateutil.tz.tzutc())
                        .timestamp(),
                    )

                    path_builder = BOX_DIR
                    oauth = setup_remote_oauth(r_c, conf=conf_obj)
                    client = Client(oauth)
                    for entry in (
                        client.file(item.object_id)
                        .get(fields=["path_collection"])
                        .path_collection["entries"]
                    ):
                        if entry.id == "0":
                            continue
                        path_builder /= entry.name
                        folder_entry = client.folder(entry.id).get(
                            fields=["modified_at"]
                        )
                        r_c.set(
                            local_or_box_file_m_time_key_func(path_builder, True),
                            parse(folder_entry.modified_at)
                            .astimezone(dateutil.tz.tzutc())
                            .timestamp(),
                        )

            break
        except BoxAPIException as e:
            crate_logger.info(f"{args}", exc_info=True)
            if e.status == 400 and e.code == "item_name_invalid":
                crate_logger.warning(
                    f"darn, bad name on the file (box does not like), {args}, cannot upload.",
                    exc_info=True,
                )
                break
            if e.status == 409:
                crate_logger.warning(
                    f"Apparently Box says this item already exists..."
                    f"and we were trying to create it. "
                    f"Need to handle this better. message: {e.message}",
                    exc_info=True,
                )
                break
        except (ConnectionError, BrokenPipeError, ProtocolError, ConnectionResetError):
            sleep_time = random.randint(0, 8)
            time.sleep(sleep_time)
            crate_logger.debug(f"{args}", exc_info=True)
            if x >= num_retries - 1:
                crate_logger.warning(
                    f"Upload giving up on: {callable_up}", exc_info=True
                )
                # no immediate plans to do anything with this info, yet.
                uploads_given_up_on.append(callable_up)
        except (TypeError, FileNotFoundError):
            crate_logger.debug("Error occurred", exc_info=True)
            break


def download_queue_processor(queue_item: "DownloadQueueItem"):
    """
    Implements a simple re-try mechanism for pending downloads
    :return:
    """
    download_queue_item = queue_item
    item = download_queue_item.item
    path = download_queue_item.path
    event = download_queue_item.event
    if event and r_c.exists(f"diy_crate:event_ids:{event.event_id}"):
        crate_logger.debug(f"Skipping download due to processed event ID for {path=}")
    if item["type"] == "file":
        info = redis_get(r_c, item) if r_c.exists(redis_key(item.object_id)) else None
        # client = Client(oauth)  # keep it around for easy access
        # hack because we did not use to store the file_path,
        # but do not want to force a download
        if info and "file_path" not in info:
            crate_logger.debug(
                f"updating local cache so the data is accurate, "
                f"but not doing the download {path=}"
            )
            info["file_path"] = path
            r_c.set(redis_key(item.object_id), json.dumps(info))
            r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
        if info and info["etag"] == item["etag"] and os.path.exists(path):
            crate_logger.debug(
                f"ETAG {info['etag']} for {path=} is identical and file exists locally, "
                f"will skip download."
            )
        # no version, or diff version, or the file does not exist locally
        if not info or info["etag"] != item["etag"] or not os.path.exists(path):
            if info:
                crate_logger.debug(
                    f"Preprocessing logic before download attempt. "
                    f"{path=} ETAG {info['etag']=} vs {item['etag']=}"
                )
            try:
                crate_logger.debug(f"attempt the download {path=}")
                perform_download(item, path)
            except (ConnectionResetError, ConnectionError):
                crate_logger.debug("Error occurred.", exc_info=True)
                time.sleep(5)
    else:
        crate_logger.debug(f"was not a file, making this a no-op on {path=}")
    if event:
        r_c.setex(
            f"diy_crate:event_ids:{event.event_id}",
            timedelta(days=32),
            (path.as_posix() if isinstance(path, Path) else path),
        )


def perform_download(item: File, path: Union[str, Path], retry_limit=15):
    if isinstance(path, str):
        path = Path(path)
    for i in range(retry_limit):
        if os.path.basename(path).startswith(".~lock"):  # avoid downloading lock files
            break
        try:
            with open(path, "wb") as item_handler:
                crate_logger.debug(
                    f"About to download: {item['name']}, " f"{item['id']}"
                )

                dlwd_key = f"diy_crate.breadcrumb.create_from_box.{path}"
                r_c.setex(dlwd_key, 300, 1)
                item.download_to(item_handler)
                # if item_wd is not None:
                #     wm.update_watch(item_wd, mask=mask | in_create)

            crate_logger.debug(f"Did download: {item['name']}, " f"{item['id']}")

        except BoxAPIException as e:
            crate_logger.info("Error occurred", exc_info=True)
            if e.status == 404:
                crate_logger.debug(
                    f"Apparently item: {item['id']}, {path} has been deleted, "
                    f"right before we tried to download",
                    exc_info=True,
                )
            elif e.status == 429:
                sleep_time = random.randint(2, pow(2, 5))
                crate_logger.info(
                    f"Rate limited during download operation with "
                    f"item: {item['id']}, "
                    f"{path}. Will sleep {sleep_time} before retry."
                )
                time.sleep(sleep_time)
                continue
            break
        else:
            if i:
                crate_logger.info(f"Retry recovered, for path: {path}")
            # path_to_add = os.path.dirname(path)
            # wm.add_watch(path=path_to_add, mask=mask, rec=True, auto_add=True)
            notify_user_with_gui(
                "Downloaded:", path.as_posix() if isinstance(path, Path) else path
            )
            was_versioned = r_c.exists(redis_key(item.object_id))
            redis_set(
                r_c,
                item,
                datetime.fromtimestamp(os.path.getmtime(path))
                .astimezone(dateutil.tz.tzutc())
                .timestamp(),
                box_dir_path=BOX_DIR,
                fresh_download=not was_versioned,
                folder=os.path.dirname(path),
            )
            time_data_map = {
                Path(path)
                .resolve()
                .as_posix(): datetime.fromtimestamp(os.path.getmtime(path))
                .astimezone(dateutil.tz.tzutc())
                .timestamp()
            }
            for mkey, mvalue in time_data_map.items():
                r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
            crate_logger.debug(f"downloaded {item}, modified_at: {item.modified_at}")
            r_c.set(
                local_or_box_file_m_time_key_func(path / item.name, True),
                parse(item.modified_at).astimezone(dateutil.tz.tzutc()).timestamp(),
            )
            path_builder = BOX_DIR
            oauth = setup_remote_oauth(r_c, conf=conf_obj)
            client = Client(oauth)
            try:
                box_file = client.file(item.object_id).get(fields=["path_collection"])
                for entry in box_file.path_collection["entries"]:
                    if entry.id == "0":
                        continue
                    path_builder /= entry.name
                    folder_entry = client.folder(entry.id).get(fields=["modified_at"])
                    r_c.set(
                        local_or_box_file_m_time_key_func(path_builder, True),
                        parse(folder_entry.modified_at)
                        .astimezone(dateutil.tz.tzutc())
                        .timestamp(),
                    )
            except boxsdk.exception.BoxAPIException as box_exc:
                if box_exc.status == 404 and box_exc.code == "trashed":
                    crate_logger.debug(
                        f"Object {item.object_id=} was previously deleted from Box. "
                        f"{(path / item.name)=}"
                    )
                    r_c.delete(redis_key(item.object_id))
                    r_c.set("diy_crate.last_save_time_stamp", int(time.time()))
                else:
                    raise
            break


# cannot use the concurrent future executor classes due to a daemon issue 3.9+.
download_pool_executor = multiprocessing.pool.ThreadPool()
upload_pool_executor = multiprocessing.pool.ThreadPool()

conf_obj = configparser.ConfigParser()
conf_dir = Path("~/.config/diycrate").expanduser().resolve()
if not conf_dir.is_dir():
    os.mkdir(conf_dir)
cloud_credentials_file_path = conf_dir / "box.ini"
if not cloud_credentials_file_path.is_file():
    cloud_credentials_file_path.write_text("")
conf_obj.read(cloud_credentials_file_path)
BOX_DIR = Path(conf_obj["box"]["directory"]).expanduser().resolve()



import logging
import subprocess
from typing import Union


def notify_user_with_gui(
    summary: str,
    body: str,
    crate_logger: Union[None, logging.Logger] = None,
    expire_time=None,
):
    """
    Sends the message to the user
    :param summary:
    :param body
    :param crate_logger:
    :param expire_time:
    :return:
    """
    notify_options = []
    # apparently notify-send does not universally use expire-time :( docs :(
    # if expire_time is not None:
    #     notify_options.append("--expire-time={expire_time}".format(expire_time=expire_time))
    notify_options.append(summary + " ")
    notify_options.append(body)
    proc = subprocess.Popen(["notify-send"] + notify_options)
    if proc.returncode:
        if crate_logger:
            crate_logger.debug(
                "Tried sending a message to user, return code: {}".format(
                    proc.returncode
                )
            )



from unittest import TestCase

from diycrate.cache_utils import redis_key


class CacheUtilTests(TestCase):
    def test_redis_key(self):
        self.assertTrue(redis_key("hello").startswith("diy_crate.version."))



import os

LOG_LEVEL = os.environ.get("DIY_CRATE_LOGGING_LEVEL")
if not LOG_LEVEL:
    LOG_LEVEL = "INFO"
LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "verbose": {
            "format": "%(levelname)s %(asctime)s %(module)s:%(lineno)d "
            "%(process)d %(thread)d %(message)s"
        },
        "simple": {"format": "%(levelname)s %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "loggers": {
        "diycrate": {"handlers": ["console"], "propagate": True, "level": LOG_LEVEL},
        "diycrate_app": {
            "handlers": ["console"],
            "propagate": True,
            "level": LOG_LEVEL,
        },
        "__main__": {"handlers": ["console"], "propagate": True, "level": LOG_LEVEL},
    },
}



import argparse
import configparser
import logging
import os
import threading
import time
from pathlib import Path
from typing import Union, Optional

import bottle
import httpx
import pyinotify
from bottle import ServerAdapter
from boxsdk import Client, exception, OAuth2
from boxsdk.auth import RemoteOAuth2
from cheroot import wsgi as wsgiserver
from cheroot.ssl.builtin import BuiltinSSLAdapter

from diycrate.cache_utils import r_c
from diycrate.file_operations import (
    EventHandler,
    wm,
    in_moved_to,
    in_moved_from,
    mask,
)

from diycrate.long_poll_processing import long_poll_event_listener
from diycrate.oauth_utils import store_tokens_callback, setup_remote_oauth, oauth_dance
from diycrate.path_utils import re_walk

from diycrate.log_utils import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)

cloud_provider_name = "Box"

bottle_app = bottle.Bottle()

# The watch manager stores the watches and provides operations on watches

# keep the lint-ing & introspection from complaining
# that these attributes don't exist before run-time.

BOX_DIR = Path()

trash_directory = Path("~/.local/share/Trash/files").expanduser()
wait_time = os.environ.get("DIY_CRATE_FILE_IO_OPERATIONS_WAIT_TIME_MS")
if wait_time:
    wait_time = int(wait_time) / 1000
else:
    wait_time = 1

handler = EventHandler(bottle_app=bottle_app, wait_time=wait_time)

file_notify_read_freq = 3

notifier = pyinotify.ThreadedNotifier(wm, handler, read_freq=file_notify_read_freq)
notifier.coalesce_events()

long_poll_thread = threading.Thread(target=long_poll_event_listener)
long_poll_thread.daemon = True

walk_thread = threading.Thread(target=re_walk)
walk_thread.daemon = True


# only re-enable when wishing to test token refresh or access dance
# @bottle_app.route('/kill_tokens')
# def kill_tokens():
#     # will get all of the oauth instances, since they are all the same reference
#     bottle_app.oauth._update_current_tokens(None, None)
#     r_c.delete('diy_crate.auth.refresh_token', 'diy_crate.auth.access_token')
#     return 'OK'

oauth: Optional[Union[RemoteOAuth2, OAuth2]] = None
conf_obj = configparser.ConfigParser()


@bottle_app.route("index")
def index():
    """
    Good to have a simple end-point.
    :return:
    """
    return "Hello, World!"


@bottle_app.route("/")
def oauth_handler():
    """
    RESTful end-point for the oauth handling
    :return:
    """
    assert bottle_app.csrf_token == bottle.request.GET["state"]
    access_token, refresh_token = httpx.post(
        conf_obj["box"]["authenticate_url"],
        data={"code": bottle.request.GET["code"]},
        verify=True,
    ).json()
    store_tokens_callback(access_token, refresh_token)
    bottle_app.oauth._update_current_tokens(access_token, refresh_token)
    if not getattr(bottle_app, "started_cloud_threads", False):
        start_cloud_threads(bottle_app.oauth)
        bottle_app.started_cloud_threads = True
    return "OK"


def start_cloud_threads(client_oauth):
    """

    :param client_oauth:
    :return:
    """
    client = Client(client_oauth)
    handler.oauth = client_oauth
    bottle_app.oauth = client_oauth
    wm.add_watch(BOX_DIR.as_posix(), mask, rec=True, auto_add=True)
    client.auth._access_token = r_c.get("diy_crate.auth.access_token")
    client.auth._refresh_token = r_c.get("diy_crate.auth.refresh_token")
    if client.auth._access_token:
        client.auth._access_token = (
            client.auth._access_token.decode(encoding="utf-8")
            if isinstance(client.auth._access_token, bytes)
            else client.auth._access_token
        )
    if client.auth._refresh_token:
        client.auth._refresh_token = (
            client.auth._refresh_token.decode(encoding="utf-8")
            if isinstance(client.auth._refresh_token, bytes)
            else client.auth._refresh_token
        )
    failed = False
    box_folder = None
    while not failed:
        try:
            box_folder = client.folder(folder_id="0").get()
        except Exception:
            crate_logger.warning(
                "Encountered error getting root box folder", exc_info=True
            )
            time.sleep(2)
        else:
            failed = True
    # local trash can
    wm.add_watch(
        trash_directory.as_posix(),
        mask=in_moved_to | in_moved_from,
        rec=True,
        auto_add=True,
    )
    if not long_poll_thread.is_alive():  # start before doing anything else
        long_poll_thread._args = (handler,)
        long_poll_thread.start()
    if not walk_thread.is_alive():
        walk_thread._args = (BOX_DIR, box_folder, client_oauth, bottle_app, handler)
        walk_thread.start()


# Create our own sub-class of Bottle's ServerAdapter
# so that we can specify SSL. Using just server='cherrypy'
# uses the default cherrypy server, which doesn't use SSL
class SSLCherryPyServer(ServerAdapter):
    """
    Custom server adapter using cherry-py with ssl
    """

    def run(self, server_handler):
        """
        Overrides super to setup Cherry py with ssl and start the server.
        :param server_handler: originating server type
        :type server_handler:
        """
        server = wsgiserver.Server((self.host, self.port), server_handler)
        # Uses the following github page's recommendation for setting up the cert:
        # https://github.com/nickbabcock/bottle-ssl
        server.ssl_adapter = BuiltinSSLAdapter(
            conf_obj["ssl"]["cacert_pem_path"], conf_obj["ssl"]["privkey_pem_path"]
        )
        try:
            server.start()
        except KeyboardInterrupt:
            walk_thread.join()
            long_poll_thread.join()
            notifier.join()
        finally:
            server.stop()


def main():
    global oauth
    global BOX_DIR

    conf_dir = Path("~/.config/diycrate").expanduser().resolve()
    if not conf_dir.is_dir():
        conf_dir.mkdir()
    cloud_credentials_file_path = conf_dir / "box.ini"
    if not cloud_credentials_file_path.is_file():
        cloud_credentials_file_path.write_text("")
    conf_obj.read(cloud_credentials_file_path)
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "--box_dir", type=str, help="directory for the root of box", default="~/box"
    )
    arg_parser.add_argument(
        "--cacert_pem_path",
        type=str,
        help="filepath to where the cacert.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--privkey_pem_path",
        type=str,
        help="filepath to where the privkey.pem is located",
        default="",
    )
    arg_parser.add_argument(
        "--authenticate_url",
        type=str,
        help="Full URL to perform remote Oauth2 authentication",
        default="",
    )
    arg_parser.add_argument(
        "--token_url",
        type=str,
        help="Full URL to do remote Oauth2 token refresh/retrieve",
        default="",
    )
    arg_parser.add_argument(
        "--authorization_url",
        type=str,
        help="Full URL to retrieve the " "Oauth2-generated authorization url",
        default="",
    )
    arg_parser.add_argument("--port", type=int, help="local web server port")
    args = arg_parser.parse_args()
    had_oauth2 = conf_obj.has_section("oauth2")
    if had_oauth2:  # this was likely before we used the RemoteOauth2 workflow,
        # so we don't want this info hanging around
        # but if there is only the client_id defined, leave it alone
        if conf_obj["oauth2"].keys() - {"client_id"}:
            conf_obj.remove_section("oauth2")
    if "ssl" not in conf_obj:
        if not args.cacert_pem_path:
            raise ValueError("Need a valid cacert_pem_path")
        if not args.privkey_pem_path:
            raise ValueError("Need a valid privkey_pem_path")
        conf_obj["ssl"] = {
            "cacert_pem_path": Path(args.cacert_pem_path)
            .expanduser()
            .resolve()
            .as_posix(),
            "privkey_pem_path": Path(args.privkey_pem_path)
            .expanduser()
            .resolve()
            .as_posix(),
        }
    conf_obj["ssl"] = {
        "cacert_pem_path": Path(args.cacert_pem_path).expanduser().resolve().as_posix()
        if args.cacert_pem_path
        else conf_obj["ssl"]["cacert_pem_path"],
        "privkey_pem_path": Path(args.privkey_pem_path)
        .expanduser()
        .resolve()
        .as_posix()
        if args.privkey_pem_path
        else conf_obj["ssl"]["privkey_pem_path"],
    }

    with open(cloud_credentials_file_path, "w") as fh:
        conf_obj.write(fh)
    had_box = conf_obj.has_section("box")
    if not had_box:
        conf_obj.add_section("box")

    conf_obj["box"] = {
        "directory": args.box_dir or conf_obj["box"]["directory"],
        "authenticate_url": args.authenticate_url
        or conf_obj["box"].get(
            "authenticate_url", "https://localhost:8081/authenticate"
        ),
        "authorization_url": args.authorization_url
        or conf_obj["box"].get("authorization_url", "https://localhost:8081/auth_url"),
        "token_url": args.token_url
        or conf_obj["box"].get("token_url", "https://localhost:8081/new_access"),
        "web_server_port": args.port or conf_obj["box"].get("web_server_port", 8080),
    }
    web_server_port = conf_obj["box"]["web_server_port"]
    bottle_thread = threading.Thread(
        target=bottle_app.run,
        kwargs=dict(server=SSLCherryPyServer, port=web_server_port, host="0.0.0.0"),
    )
    bottle_thread.daemon = True
    with cloud_credentials_file_path.open("w") as fh:
        conf_obj.write(fh)
    BOX_DIR = Path(conf_obj["box"]["directory"]).expanduser().resolve()
    if not BOX_DIR.is_dir():
        BOX_DIR.mkdir()
    if not (
        r_c.exists("diy_crate.auth.access_token")
        and r_c.exists("diy_crate.auth.refresh_token")
    ):
        oauth_dance(
            redis_client=r_c,
            conf=conf_obj,
            bottle_app=bottle_app,
            file_event_handler=handler,
        )
        oauth = bottle_app.oauth
    else:
        try:
            oauth = setup_remote_oauth(r_c, conf=conf_obj)
            start_cloud_threads(oauth)
            bottle_app.started_cloud_threads = True
        except exception.BoxOAuthException:
            r_c.delete("diy_crate.auth.access_token", "diy_crate.auth.refresh_token")
            oauth_dance(
                redis_client=r_c,
                conf=conf_obj,
                bottle_app=bottle_app,
                file_event_handler=handler,
            )
    notifier.start()
    # notifier_thread = threading.Thread(target=notifier.loop)
    # notifier_thread.daemon = True
    # notifier_thread.start()
    if bottle_thread and not bottle_thread.is_alive():
        bottle_thread.start()


if __name__ == "__main__":
    main()



import webbrowser
from configparser import ConfigParser
from typing import Union

import httpx

from boxsdk import OAuth2
from boxsdk.auth import RemoteOAuth2

from .cache_utils import r_c

import logging

from .log_utils import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)


def setup_oauth(cache_client, conf_object, callback):
    """
    sets up the oauth instance with credentials and runtime callback.
    :param cache_client:
    :param conf_object:
    :param callback:
    :return:
    """
    access_token = cache_client.get("diy_crate.auth.access_token")
    refresh_token = cache_client.get("diy_crate.auth.refresh_token")
    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")
    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    oauth = OAuth2(
        client_id=conf_object["oauth2"]["client_id"],
        client_secret=conf_object["oauth2"]["client_secret"],
        store_tokens=callback,
        access_token=access_token,
        refresh_token=refresh_token,
    )
    return oauth


def get_access_token(access_token):
    """

    :param access_token: token we will try to refresh/get anew
    :return:
    """
    from .file_operations import conf_obj

    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")

    conf_object = conf_obj
    remote_url = conf_object["box"]["token_url"]
    refresh_token = r_c.get("diy_crate.auth.refresh_token")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")

    access_token, refresh_token = httpx.post(
        remote_url,
        data={"refresh_token": refresh_token, "access_token": access_token},
        verify=True,
    ).json()
    r_c.set("diy_crate.auth.access_token", access_token)
    r_c.set("diy_crate.auth.refresh_token", refresh_token)
    crate_logger.debug(
        f"new access_token: {access_token}, refresh_token: {refresh_token}"
    )
    return access_token


def setup_remote_oauth(cache_client, retrieve_access_token=get_access_token, conf=None):
    """
    sets up the oauth instance with credentials and runtime callback.
    :param cache_client:
    :param retrieve_access_token:
    :param conf:
    :return:
    """
    conf: Union[ConfigParser, dict] = conf or {}
    access_token = cache_client.get("diy_crate.auth.access_token")

    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")
    refresh_token = cache_client.get("diy_crate.auth.refresh_token")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    if isinstance(conf, ConfigParser):
        client_id = (
            (conf["oauth2"]["client_id"] or "") if conf.has_section("oauth2") else ""
        )
    else:
        # raise Exception
        client_id = conf.get("oauth2", {}).get("client_id", "")
    oauth = RemoteOAuth2(
        client_id=client_id,
        client_secret="",
        access_token=access_token,
        refresh_token=refresh_token,
        retrieve_access_token=retrieve_access_token,
    )
    return oauth


def store_tokens_callback(access_token, refresh_token):
    """
    Intention is to store the oauth tokens
    :param access_token:
    :param refresh_token:
    :return:
    """
    assert access_token
    assert refresh_token
    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")

    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    r_c.set("diy_crate.auth.access_token", access_token)
    r_c.set("diy_crate.auth.refresh_token", refresh_token)


def oauth_dance(redis_client, conf, bottle_app, file_event_handler, bottle_thread=None):
    bottle_app.oauth = file_event_handler.oauth = setup_remote_oauth(
        redis_client, conf=conf
    )
    redirect_url_domain = "127.0.0.1"
    (auth_url, csrf_token) = bottle_app.oauth.get_authorization_url(
        redirect_url=f"https://{redirect_url_domain}:{conf['box']['web_server_port']}/"
    )
    print(auth_url, csrf_token)
    bottle_app.csrf_token = csrf_token
    if bottle_thread and not bottle_thread.is_alive():
        bottle_thread.start()
    webbrowser.open_new_tab(auth_url)  # make it easy for the end-user to start auth



import os
import time
import logging
from datetime import timedelta, datetime
from functools import partial
from os import PathLike
from pathlib import Path
from typing import Union, Optional, Dict, List

from dateutil.tz import tzutc
from bottle import Bottle
from boxsdk import OAuth2
from boxsdk.exception import BoxAPIException
from boxsdk.client import Client
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from dateutil.parser import parse

from .file_operations import wm, BOX_DIR, path_time_recurse_func
from .item_queue_io import (
    DownloadQueueItem,
    UploadQueueItem,
    download_pool_executor,
    download_queue_processor,
    upload_pool_executor,
    upload_queue_processor,
)
from .cache_utils import (
    redis_key,
    r_c,
    redis_set,
    redis_get,
    local_or_box_file_m_time_key_func,
)
from .iter_utils import SafeIter
from .log_utils import setup_logger

setup_logger()

crate_logger = logging.getLogger(__name__)


def walk_and_notify_and_download_tree(
    path: PathLike,
    box_folder: Folder,
    oauth_obj: OAuth2,
    bottle_app: Bottle,
    p_id: Optional[str] = None,
    local_only: bool = False,
) -> None:
    """
    Walk the path recursively and add watcher and create the path.
    :param path:
    :param box_folder:
    :param oauth_obj:
    :param p_id:
    :param bottle_app:
    :param file_event_handler:
    :param local_only:
    :return:
    """
    start_t = time.monotonic()
    if os.path.isdir(path):
        # wm.add_watch(path, mask, rec=True, auto_add=True)
        local_files = os.listdir(path)
    else:
        raise ValueError(f"path: {path} is not a path; cannot walk it.")
    path = Path(path)
    crate_logger.debug(f"Walking path: {path}")
    bottle_app.oauth = getattr(bottle_app, "oauth", oauth_obj)
    client = oauth_setup_within_directory_walk(bottle_app.oauth)

    time_data_map = path_time_recurse_func(path, wm)
    if Path(path) == BOX_DIR:
        for mkey, mvalue in time_data_map.items():
            r_c.set(local_or_box_file_m_time_key_func(mkey, False), mvalue)
    while True:
        try:
            basic_folder_keys = ["id", "etag", "name", "path_collection", "modified_at"]
            b_folder: Folder = client.folder(folder_id=box_folder.object_id).get(
                fields=basic_folder_keys
            )
            break
        except ConnectionError:
            crate_logger.warning("Trying again after small sleep", exc_info=True)
            time.sleep(5)

    folder_cache_info = (
        redis_get(r_c, b_folder) if r_c.exists(redis_key(b_folder.object_id)) else None
    )
    do_process = True
    if folder_cache_info and not local_only:
        do_process = any_unresolved_modifications(b_folder, path, time_data_map)
    if not do_process:
        return
    limit = 100
    for _ in range(5):
        try:
            if not local_only:
                local_files_walk_pre_process(
                    b_folder, limit, local_files, oauth_obj, path
                )
            break
        except Exception:
            pass

    ids_in_folder = []
    offset = 0
    if not local_only:
        folder_items = b_folder.get_items(
            limit=limit, offset=offset, fields=["name", "id", "etag", "modified_at"]
        )
    else:
        folder_items = iter([])
    safe_iter = SafeIter(folder_items)
    for box_item in safe_iter:
        if box_item is None:
            continue
        if box_item.object_id not in ids_in_folder:
            ids_in_folder.append(box_item.object_id)
        if box_item["name"] in local_files:
            local_files.remove(box_item["name"])
        if box_item["type"] == "folder":
            local_path = path / box_item["name"]
            fresh_download = False
            if not os.path.isdir(local_path):
                os.mkdir(local_path)
                fresh_download = True
            retry_limit = 15
            kick_off_sub_directory_box_folder_download_walk(
                bottle_app,
                box_folder,
                box_item,
                client,
                fresh_download,
                local_path,
                oauth_obj,
                retry_limit,
            )
        else:
            kick_off_download_file_from_box_via_walk(box_item, oauth_obj, path)
        r_c.set(
            local_or_box_file_m_time_key_func(path / box_item.name, True),
            parse(box_item.modified_at).astimezone(tzutc()).timestamp(),
        )
    if not local_only:
        redis_set(
            cache_client=r_c,
            cloud_item=b_folder,
            last_modified_time=datetime.fromtimestamp(os.path.getmtime(path))
            .astimezone(tzutc())
            .timestamp(),
            box_dir_path=BOX_DIR,
            fresh_download=not r_c.exists(redis_key(box_folder.object_id)),
            folder=os.path.dirname(path),
            sub_ids=ids_in_folder,
            parent_id=p_id,
        )
    end_t = time.monotonic()
    crate_logger.debug(
        f"walk_and_notify_and_download_tree on {box_folder}, took: {end_t - start_t:.2f}s"
    )


def any_unresolved_modifications(
    b_folder: Folder, path: Path, time_data_map: Dict[str, float]
) -> bool:
    have_any = True
    b_folder_modified_at = b_folder["modified_at"]
    if b_folder_modified_at:
        path_sorted_by_m_time = sorted(
            (
                (Path(k).as_posix(), v)
                for k, v in time_data_map.items()
                if path in Path(k).parents or Path(k) == path
            ),
            key=lambda x: x[1],
        )
        if path_sorted_by_m_time:
            redis_path_sorted_by_m_time = sorted(
                (
                    (
                        Path(
                            k.decode()[
                                k.decode().find(BOX_DIR.as_posix()) :  # noqa: E203
                            ]  # noqa: E203
                        ).as_posix(),
                        float(r_c.get(k)),
                    )
                    for k in r_c.keys(
                        local_or_box_file_m_time_key_func(path, False) + "*"
                    )
                ),
                key=lambda x: x[1],
            )
            # box folder's modified_at value changes when it, a file,
            # or a sub-directory's files changes,
            # so we only care about this specific folder's key.
            box_redis_path_sorted_by_m_time = sorted(
                (
                    (
                        Path(
                            k.decode()[
                                k.decode().find(BOX_DIR.as_posix()) :  # noqa: E203
                            ]  # noqa: E203
                        ).as_posix(),
                        float(r_c.get(k)),
                    )
                    for k in r_c.keys(local_or_box_file_m_time_key_func(path, True))
                ),
                key=lambda x: x[1],
            )
            if redis_path_sorted_by_m_time and box_redis_path_sorted_by_m_time:
                # box folder's modified_at value changes when it, a file,
                # or a sub-directory's files changes,
                # so we only care about this specific folder's key.
                box_current_path_and_m_time = tuple(
                    [
                        (path.parent / b_folder.name).as_posix(),
                        parse(b_folder_modified_at).astimezone(tzutc()).timestamp(),
                    ]
                )
                differences_full = []
                differences_full.append(
                    {
                        "kind": "local",
                        "value": [
                            list(path_sorted_by_m_time[-1]),
                            list(redis_path_sorted_by_m_time[-1]),
                        ],
                    }
                )
                differences_full.append(
                    {
                        "kind": "box",
                        "value": [
                            list(box_current_path_and_m_time),
                            list(box_redis_path_sorted_by_m_time[-1]),
                        ],
                    }
                )
                differences = [
                    item
                    for item in differences_full
                    if item["value"][0] != item["value"][1]
                ]
                if not differences:
                    crate_logger.debug(
                        f"Skipping walk on folder {b_folder}; "
                        f"neither the box or local history has changed since the last run."
                    )

                    have_any = False
                else:
                    # crate_logger.debug(f"are differences: {differences}")
                    have_any = differences
    elif path == BOX_DIR:
        path_sorted_by_m_time = sorted(
            (
                (Path(k).as_posix(), v)
                for k, v in time_data_map.items()
                if path in Path(k).parents
            ),
            key=lambda x: x[1],
        )
        redis_path_sorted_by_m_time = sorted(
            (
                (
                    Path(
                        k.decode()[k.decode().find(BOX_DIR.as_posix()) :]  # noqa: E203
                    ).as_posix(),
                    float(r_c.get(k)),
                )
                for k in r_c.keys(local_or_box_file_m_time_key_func(path, False) + "*")
                if r_c.get(k)
            ),
            key=lambda x: x[1],
        )
        box_redis_path_sorted_by_m_time = sorted(
            (
                (
                    Path(
                        k.decode()[
                            k.decode().find(BOX_DIR.as_posix()) :  # noqa: E203
                        ]  # noqa: E203
                    ).as_posix(),
                    float(r_c.get(k)),
                )
                for k in r_c.keys(local_or_box_file_m_time_key_func(path, True) + "*")
                if r_c.get(k)
            ),
            key=lambda x: x[1],
        )

        b_children = sorted(
            b_folder.get_items(fields=["name", "modified_at", "path_collection"]),
            key=lambda x: datetime.fromisoformat(x.modified_at)
            .astimezone(tzutc())
            .timestamp(),
        )
        b_oldest_child: Optional[Union[Folder, File]] = (
            b_children[-1] if b_children else None
        )
        if b_oldest_child:
            crate_logger.debug([item.name for item in b_children])
        b_oldest_child_modified_at = (
            datetime.fromisoformat(b_oldest_child.modified_at)
            .astimezone(tzutc())
            .timestamp()
            if b_oldest_child
            else None
        )
        if box_redis_path_sorted_by_m_time:
            # because of the keys pattern, we received back all paths which matched path*,
            # but we really only want path | path/*/
            # that is, our matching path, or just its first level files & sub directories.
            box_redis_path_sorted_by_m_time = [
                item
                for item in box_redis_path_sorted_by_m_time
                if Path(item[0]).parent == path or Path(item[0]) == path
            ]

        if (
            path_sorted_by_m_time
            and redis_path_sorted_by_m_time
            and box_redis_path_sorted_by_m_time
            and b_oldest_child_modified_at
            and b_oldest_child
        ):
            differences_full: List[
                Dict[str, Union[str, List[List[Union[str, float]]]]]
            ] = [
                {
                    "kind": "local",
                    "value": [
                        list(path_sorted_by_m_time[-1]),
                        list(redis_path_sorted_by_m_time[-1]),
                    ],
                }
            ]
            box_current_path_of_choice = (
                path
                / "/".join(
                    entry.name
                    for entry in b_oldest_child.path_collection["entries"]
                    if entry.id != "0"
                )
                / b_oldest_child.name
            ).as_posix()
            box_redis_entry_of_choice = list(box_redis_path_sorted_by_m_time[-1])
            differences_full.append(
                {
                    "kind": "box",
                    "value": [
                        [box_current_path_of_choice, b_oldest_child_modified_at],
                        box_redis_entry_of_choice,
                    ],
                }
            )
            differences = [
                item
                for item in differences_full
                if item["value"][0] != item["value"][1]
            ]
            if not differences:
                crate_logger.debug(
                    f"folder {b_folder} on "
                    f"{path.as_posix()} has not been changed locally, "
                    f"new local vs prev redis local ({path_sorted_by_m_time[-1]}, "
                    f"{redis_path_sorted_by_m_time[-1]}), "
                    f"and also current newest box child against new box child "
                    f"in redis, we will skip!"
                )
                have_any = False
            else:
                crate_logger.debug(
                    f"Difference detected on {path.as_posix()}, {differences}"
                )
                have_any = differences
    return have_any


def oauth_setup_within_directory_walk(oauth):
    client = Client(oauth)
    access_token = r_c.get("diy_crate.auth.access_token")
    if isinstance(access_token, bytes):
        access_token = access_token.decode(encoding="utf-8")
    refresh_token = r_c.get("diy_crate.auth.refresh_token")
    if isinstance(refresh_token, bytes):
        refresh_token = refresh_token.decode(encoding="utf-8")
    if access_token:
        client.auth._access_token = access_token
    if refresh_token:
        client.auth._refresh_token = refresh_token
    return client


def kick_off_download_file_from_box_via_walk(box_item, oauth_obj, path):
    try:
        file_obj = box_item
        queue_item = DownloadQueueItem(
            file_obj, os.path.join(path, box_item["name"]), oauth_obj, None
        )
        download_pool_executor.apply_async(
            download_queue_processor, kwds=dict(queue_item=queue_item)
        )

    except BoxAPIException as e:
        crate_logger.debug("Error occurred", exc_info=True)
        if e.status == 404:
            crate_logger.debug(
                f"Box says: {box_item['id']}, {box_item['name']}, is a 404 status."
            )
            if r_c.exists(redis_key(box_item.object_id)):
                crate_logger.debug(f"Deleting {box_item['id']}, {box_item['name']}")
                r_c.delete(redis_key(box_item.object_id))
        raise e


def kick_off_sub_directory_box_folder_download_walk(
    bottle_app: Bottle,
    box_folder: Folder,
    box_item: Union[File, Folder],
    client: Client,
    fresh_download: bool,
    local_path: PathLike,
    oauth_obj: OAuth2,
    retry_limit: int,
) -> None:
    bottle_app.oauth = getattr(bottle_app, "oauth", oauth_obj)
    start_t = time.monotonic()
    for i in range(0, retry_limit):
        try:
            redis_set(
                cache_client=r_c,
                cloud_item=box_item,
                last_modified_time=datetime.fromtimestamp(os.path.getmtime(local_path))
                .astimezone(tzutc())
                .timestamp(),
                box_dir_path=BOX_DIR,
                fresh_download=fresh_download,
                folder=os.path.dirname(local_path),
            )
            box_folder_obj = client.folder(folder_id=box_item.object_id)
        except BoxAPIException as e:
            crate_logger.debug("Box error occurred.")
            if e.status == 404:
                crate_logger.debug(
                    f"Box says: {box_item['id']}, {box_item['name']}, is a 404 status."
                )
                crate_logger.debug(
                    "But, this is a folder, we do not handle recursive "
                    "folder deletes correctly yet."
                )
                break
        except (ConnectionError, ConnectionResetError, BrokenPipeError, OSError):
            crate_logger.debug(f"Attempt {i + 1}/{retry_limit}", exc_info=True)
        else:
            if i:
                crate_logger.debug("Succeeded on retry.")
            walk_and_notify_and_download_tree(
                local_path,
                box_folder_obj,
                oauth_obj,
                bottle_app,
                p_id=box_folder.object_id,
            )
            break
    end_t = time.monotonic()
    crate_logger.debug(
        f"kick_off_sub_directory_box_folder_download_walk "
        f"on {box_item}, took: {end_t - start_t:.2f}s"
    )


def local_files_walk_pre_process(
    b_folder: Folder, limit, local_files, oauth_obj, path, offset=0
):
    folder_items = b_folder.get_items(
        limit=limit, offset=offset, fields=["name", "etag", "id", "modified_at"]
    )
    safe_iter = SafeIter(folder_items)
    start_t = time.monotonic()
    for box_item in safe_iter:
        if box_item is None:
            continue
        if box_item["name"] in local_files:
            local_files.remove(box_item["name"])
    for (
        local_file
    ) in local_files:  # prioritize the local_files not yet on box's server.
        cur_box_folder = b_folder
        local_path = os.path.join(path, local_file)
        if os.path.isfile(local_path):
            crate_logger.info(f"upload, {local_path}, {local_file}")
            m_timestamp = (
                datetime.fromtimestamp(os.path.getmtime(local_path))
                .astimezone(tzutc())
                .timestamp()
            )

            queue_item = UploadQueueItem(
                m_timestamp,
                partial(cur_box_folder.upload, local_path, local_file),
                oauth_obj,
                local_path,
            )
            # upload_queue.put(
            #     queue_item
            # )
            upload_pool_executor.apply_async(
                upload_queue_processor, kwds=dict(queue_item=queue_item)
            )

    end_t = time.monotonic()
    crate_logger.debug(
        f"local_files_walk_pre_process, {b_folder}, took: {end_t - start_t:.2f}s"
    )


def re_walk(path, box_folder, oauth_obj, bottle_app: Bottle, file_event_handler=None):
    """

    :param path:
    :param box_folder:
    :param oauth_obj:
    :param bottle_app:
    :param file_event_handler:
    :return:
    """
    while True:
        try:
            start = time.time()
            crate_logger.info("Starting walk.")
            # wm.add_watch(path.as_posix(), mask, rec=True, auto_add=True)
            walk_and_notify_and_download_tree(path, box_folder, oauth_obj, bottle_app)
            end = time.time()
            duration: Union[int, float] = int(end - start)
            if duration >= 60:
                duration = round(timedelta(seconds=duration) / timedelta(minutes=1), 2)
                unit = "m"
            else:
                unit = "s"
            crate_logger.info(f"Finished walking! Took {duration}{unit}")
            time.sleep(3600)  # once an hour we walk the tree
        except KeyboardInterrupt:
            raise
        except Exception:
            crate_logger.warning(
                "re_walk encountered error (will reloop).", exc_info=True
            )



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

install_cmd = None
try:
    from install_cmd import InstallWrapper

    install_cmd = InstallWrapper
except ModuleNotFoundError:
    pass

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")


extra = dict()

if install_cmd:
    cmdclass = {"install": install_cmd}
    extra["cmdclass"] = cmdclass

setup(
    url="https://github.com/jheld/diycrate",
    test_suite="tests",
    packages=find_packages(exclude=["tests", "tests.*"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    **extra,
)
